from typing import Optional, List, Dict, Mapping
from dataclasses import dataclass

from annotell.input_api.util import filter_none


@dataclass
class CameraSettings:
    width: int
    height: int

    def to_dict(self):
        return filter_none({
            "width": self.width,
            "height": self.height
        })


@dataclass
class SensorSpecification:
    sensor_to_pretty_name: Optional[Dict[str, str]] = None
    sensor_order: Optional[List[str]] = None

    def to_dict(self):
        return filter_none({
            "sensorToPrettyName": self.sensor_to_pretty_name,
            "sensorOrder": self.sensor_order,
        })



from dataclasses import dataclass
from typing import List, Dict, Optional

from annotell.input_api.model.calibration import Position, RotationQuaternion


@dataclass
class EgoVehiclePose:
    """Both `position` and `rotation` are with respect to the local coordinate system (LCS)."""
    position: Position
    rotation: RotationQuaternion

    def to_dict(self) -> Dict:
        return dict(
            position=self.position.to_dict(),
            rotation=self.rotation.to_dict()
        )



import logging
from typing import Optional, Dict, List

import annotell.input_api.model as IAM
from annotell.input_api.file_resource_client import FileResourceClient
from annotell.input_api.http_client import HttpClient
from annotell.input_api.util import get_image_dimensions

log = logging.getLogger(__name__)


class CreateableInputAPIResource(FileResourceClient):

    def __init__(self, client: HttpClient, file_resource_client: FileResourceClient):
        super().__init__()
        self._client = client
        self._file_resource_client = file_resource_client

    def _post_input_request(self, resource_path: str,
                            input_request: dict,
                            project: Optional[str],
                            batch: Optional[str],
                            annotation_types: Optional[List[str]] = None,
                            dryrun: bool = False) -> Optional[IAM.InputJobCreated]:
        """
        Send input to Input API. if not dryrun is true, only validation is performed
        Otherwise, returns `InputJobCreated`
        """
        if annotation_types:
            input_request['annotationTypes'] = annotation_types

        log.debug("POST:ing to %s input %s", resource_path, input_request)

        request_url = self._resolve_request_url(resource_path, project, batch)
        json_resp = self._client.post(
            request_url, json=input_request, dryrun=dryrun)
        if not dryrun:
            response = IAM.InputJobCreated.from_json(json_resp)

            if (len(response.files) > 0):
                self._file_resource_client.upload_files(response.files)
                self._client.post(
                    f"v1/inputs/input-jobs/{response.input_uuid}/commit",
                    json=False,
                    discard_response=True
                )

            return response

    @staticmethod
    def _resolve_request_url(resource_path: str,
                             project: Optional[str] = None,
                             batch: Optional[str] = None) -> str:
        """
        Resolves which request url to use for input based on if project and batch is specified
        """
        url = f"v1/inputs/"

        if project is not None:
            url += f"project/{project}/"
            if batch is not None:
                url += f"batch/{batch}/"

        url += resource_path

        return url



from dataclasses import dataclass
from annotell.input_api.model.abstract.abstract_models import Response
from typing import Union


@dataclass
class Annotation(Response):
    annotation_id: int
    export_content: Union[list, dict]

    @staticmethod
    def from_json(js: dict):
        return Annotation(
            annotation_id=int(js["annotationId"]),
            export_content=js["exportContent"]
        )



from typing import List
from dataclasses import dataclass, field
from annotell.input_api.model.input.resources.image import Image
from annotell.input_api.model.input.resources.point_cloud import PointCloud


@dataclass
class Frame:
    point_clouds: List[PointCloud] = field(default_factory=list)
    images: List[Image] = field(default_factory=list)

    def to_dict(self) -> dict:
        return dict(
            pointClouds=[pc.to_dict() for pc in self.point_clouds] if self.point_clouds else None,
            images=[image.to_dict() for image in self.images] if self.images else None)



from annotell.input_api.model.abstract.abstract_models import Response
from annotell.input_api.util import ts_to_dt
from datetime import datetime
from typing import Mapping, Optional
from dataclasses import dataclass
from typing import Dict, Union
from annotell.input_api.model.calibration.sensors import (
    CameraCalibration, LidarCalibration
)


@dataclass
class SensorCalibration:
    external_id: str
    calibration: Dict[str, Union[CameraCalibration, LidarCalibration]]

    def to_dict(self):
        return {
            'externalId': self.external_id,
            'calibration': {k: v.to_dict() for (k, v) in self.calibration.items()}
        }

@dataclass
class SensorCalibrationEntry(Response):
    id: str
    external_id: str
    created: datetime
    calibration: Optional[Mapping[str, Union[CameraCalibration, LidarCalibration]]]

    @staticmethod
    def from_json(js: dict):
        return SensorCalibrationEntry(
            id=js["id"],
            external_id=js["externalId"],
            created=ts_to_dt(js["created"]),
            calibration=js.get("calibration")
        )



from dataclasses import dataclass
from typing import Dict
from annotell.input_api.model.abstract.abstract_models import Response


@dataclass
class InputJobCreated(Response):
    input_uuid: str
    files: Dict[str, str]

    @staticmethod
    def from_json(js: dict):
        return InputJobCreated(js["internalId"], js["files"])

    def __str__(self):
        return f"{self.__class__.__name__}(input_uuid={self.input_uuid}, files={{...}})"


@dataclass
class CreateInputResponse:
    input_uuid: str

    @staticmethod
    def from_input_job_response(input_job: InputJobCreated):
        return CreateInputResponse(input_job.input_uuid)



from typing import List, Dict

import annotell.input_api.model.annotation as AnnotationModel
from annotell.input_api.util import filter_none
from annotell.input_api.resources.abstract import InputAPIResource


class AnnotationResource(InputAPIResource):
    def get_annotations(
        self,
        input_uuids: List[str]
    ) -> Dict[str, List[AnnotationModel.Annotation]]:
        """
        Returns the export ready annotations, either
        * All annotations connected to a specific request, if a request id is given
        * All annotations connected to the organization of the user, if no request id is given

        :param input_uuids: List with input uuid
        :param request_id: An id of a request
        :return Dict: A dictionary containing the ready annotations
        """
        external_id_query_param = ",".join(input_uuids) if input_uuids else None
        json_resp = self._client.get("v1/annotations", params=filter_none({
            "inputs": external_id_query_param
        }))


        annotations = dict()
        for k, v in json_resp.items():
            annotations[k] = [
                AnnotationModel.Annotation.from_json(annotation) for annotation in v
            ]
        return annotations



"""Client for communicating with the Annotell platform."""
import logging
import random
import time
from pathlib import Path
from typing import Mapping, Dict, BinaryIO, Optional

import requests

from annotell.input_api.util import get_content_type

log = logging.getLogger(__name__)

RETRYABLE_STATUS_CODES = [408, 429, 500, 501, 502, 503, 504, 505, 506, 507, 508, 509, 510, 511, 598, 599]


class FileResourceClient:

    def __init__(self,
                 max_upload_retry_attempts: int = 23,
                 max_upload_retry_wait_time: int = 60):
        """
        :param max_upload_retry_attempts: Max number of attempts to retry uploading a file to GCS.
        :param max_upload_retry_wait_time:  Max with time before retrying an upload to GCS.
        """
        self.MAX_NUM_UPLOAD_RETRIES = max_upload_retry_attempts
        self.MAX_RETRY_WAIT_TIME = max_upload_retry_wait_time  # seconds

    def _get_wait_time(self, upload_attempt: int) -> int:
        """
        Calculates the wait time before attempting another file upload to GCS

        :param upload_attempt: How many attempts to upload that have been made
        :return: int: The time to wait before retrying upload
        """
        max_wait_time = pow(2, upload_attempt - 1)
        wait_time = random.random() * max_wait_time
        wait_time = wait_time if wait_time < self.MAX_RETRY_WAIT_TIME else self.MAX_RETRY_WAIT_TIME
        return wait_time

    #  Using similar retry strategy as gsutil
    #  https://cloud.google.com/storage/docs/gsutil/addlhelp/RetryHandlingStrategy
    def _upload_file(self, upload_url: str, file: BinaryIO, headers: Dict[str, str], upload_attempt: int = 1) -> None:
        """
        Upload the file to GCS, retries if the upload fails with some specific status codes.
        """
        log.info(f"Uploading file={file.name}")
        resp = requests.put(upload_url, data=file, headers=headers)
        try:
            resp.raise_for_status()
        except requests.HTTPError as e:
            log.error(f"On upload attempt ({upload_attempt}/{self.MAX_NUM_UPLOAD_RETRIES}) to GCS "
                      f"got response:\n{resp.status_code}: {resp.content}")

            if upload_attempt < self.MAX_NUM_UPLOAD_RETRIES and resp.status_code in RETRYABLE_STATUS_CODES:
                wait_time = self._get_wait_time(upload_attempt)
                log.info(f"Waiting {int(wait_time)} seconds before retrying")
                time.sleep(wait_time)
                self._upload_file(upload_url, file, headers, upload_attempt + 1)
            else:
                raise e

        except Exception as e:
            raise e

    def upload_files(self, url_map: Mapping[str, str], folder: Optional[Path] = None) -> None:
        """
        Upload all files to cloud storage

        :param url_map: map between filename and GCS signed URL
        :param folder: Optional base path, will join folder and each filename in map if provided
        """
        for (filename, upload_url) in url_map.items():
            file_path = folder.joinpath(filename).expanduser() if folder else Path(filename).expanduser()
            with file_path.open('rb') as file:
                content_type = get_content_type(filename)
                headers = {"Content-Type": content_type}
                self._upload_file(upload_url, file, headers)



from dataclasses import dataclass, field
from typing import List, Mapping, Union

from annotell.input_api.model.input.abstract.sequence_frame import SequenceFrame
from annotell.input_api.model.input.resources import Image, VideoFrame


@dataclass
class Frame(SequenceFrame):
    images: List[Image] = field(default_factory=list)
    video_frames: List[VideoFrame] = field(default_factory=list)
    metadata: Mapping[str, Union[int, float, str, bool]] = field(default_factory=dict)

    def to_dict(self) -> dict:
        return dict(
            frameId=self.frame_id,
            relativeTimestamp=self.relative_timestamp,
            images=[image.to_dict() for image in self.images] if self.images else None,
            videoFrames=[vf.to_dict() for vf in self.video_frames] if self.video_frames else None,
            metadata=self.metadata
        )



"""Client for communicating with the Annotell platform."""
import logging

from annotell.auth.authsession import DEFAULT_HOST as DEFAULT_AUTH_HOST

from annotell.input_api.file_resource_client import FileResourceClient
from annotell.input_api.http_client import HttpClient
from annotell.input_api.resources.annotation.annotation import AnnotationResource
from annotell.input_api.resources.calibration.calibration import CalibrationResource
from annotell.input_api.resources.input.cameras import Cameras
from annotell.input_api.resources.input.cameras_sequence import CamerasSequence
from annotell.input_api.resources.input.input import InputResource
from annotell.input_api.resources.input.lidars_and_cameras import LidarsAndCameras
from annotell.input_api.resources.input.lidars_and_cameras_sequence import LidarsAndCamerasSequence
from annotell.input_api.resources.project.project import ProjectResource

DEFAULT_HOST = "https://input.annotell.com"

log = logging.getLogger(__name__)


class InputApiClient:
    """Creates Annotell inputs from local files."""

    def __init__(self, *,
                 auth=None,
                 host: str = DEFAULT_HOST,
                 auth_host: str = DEFAULT_AUTH_HOST,
                 client_organization_id: int = None,
                 max_upload_retry_attempts: int = 23,
                 max_upload_retry_wait_time: int = 60):
        """
        :param auth: auth credentials, see https://github.com/annotell/annotell-python/tree/master/annotell-auth
        :param host: override for input api url
        :param auth_host: override for authentication url
        :param client_organization_id: Overrides your users organization id. Only works with an Annotell user.
        :param max_upload_retry_attempts: Max number of attempts to retry uploading a file to GCS.
        :param max_upload_retry_wait_time:  Max with time before retrying an upload to GCS.
        """

        self._client = HttpClient(auth=auth,
                                  host=host,
                                  auth_host=auth_host,
                                  client_organization_id=client_organization_id)
        self._file_client = FileResourceClient(max_upload_retry_attempts=max_upload_retry_attempts,
                                               max_upload_retry_wait_time=max_upload_retry_wait_time)

        self.calibration = CalibrationResource(self._client)
        self.project = ProjectResource(self._client)
        self.annotation = AnnotationResource(self._client)
        self.input = InputResource(self._client)

        self.lidar_and_cameras = LidarsAndCameras(self._client, self._file_client)
        self.lidars_and_cameras_sequence = LidarsAndCamerasSequence(self._client, self._file_client)
        self.cameras = Cameras(self._client, self._file_client)
        self.cameras_sequence = CamerasSequence(self._client, self._file_client)



from dataclasses import dataclass, field
from typing import List

from annotell.input_api.model.input.resources.image import Image


@dataclass
class Frame:
    images: List[Image] = field(default_factory=list)

    def to_dict(self) -> dict:
        return dict(images=[image.to_dict() for image in self.images] if self.images else None)



from dataclasses import dataclass
from datetime import datetime
from enum import Enum
from annotell.input_api.util import ts_to_dt
from annotell.input_api.model.abstract.abstract_models import Response


class ProjectBatchStatus(str, Enum):
    PENDING = 'pending'
    OPEN = 'open'
    READY = 'ready'
    INPROGESS = 'in-progress'
    COMPLETED = 'completed'


@dataclass
class ProjectBatch(Response):
    project: str
    batch: str
    title: str
    status: ProjectBatchStatus
    created: datetime
    updated: datetime

    @staticmethod
    def from_json(js: dict):
        return ProjectBatch(
            project=js["project"],
            batch=js["batch"],
            title=js["title"],
            status=ProjectBatchStatus(js["status"]),
            created=ts_to_dt(js["created"]),
            updated=ts_to_dt(js["updated"])
        )



"""Client for communicating with the Annotell platform."""
import logging

import requests
from annotell.auth.authsession import FaultTolerantAuthRequestSession
from annotell.input_api.util import filter_none

from annotell.input_api import __version__

log = logging.getLogger(__name__)

ENVELOPED_JSON_TAG = "data"


class HttpClient:
    """Http Client dealing with auth and communication with API."""

    def __init__(self,
                 auth,
                 host: str,
                 auth_host: str,
                 client_organization_id: int = None):
        """
        :param auth: auth credentials, see https://github.com/annotell/annotell-python/tree/master/annotell-auth
        :param host: override for input api url
        :param auth_host: override for authentication url
        :param client_organization_id: Overrides your users organization id. Only works with an Annotell user.
        :param max_upload_retry_attempts: Max number of attempts to retry uploading a file to GCS.
        :param max_upload_retry_wait_time:  Max with time before retrying an upload to GCS.
        """

        self.host = host
        self._auth_req_session = FaultTolerantAuthRequestSession(host=auth_host, auth=auth)
        self.headers = {
            "Accept-Encoding": "gzip",
            "Accept": "application/json",
            "User-Agent": "annotell-input-api/%s" % __version__
        }
        self.dryrun_header = {"X-Dryrun": ""}

        if client_organization_id is not None:
            self.headers["X-Organization-Id"] = str(client_organization_id)
            log.warning(f"WARNING: You will now act as if you are part of organization: {client_organization_id}. "
                        f"This will not work unless you are an Annotell user.")

    @property
    def session(self):
        return self._auth_req_session

    @staticmethod
    def _raise_on_error(resp: requests.Response) -> requests.Response:
        try:
            resp.raise_for_status()
        except requests.HTTPError as exception:
            if exception.response is not None and exception.response.status_code == 400:
                try:
                    message = exception.response.json()["message"]
                except ValueError:
                    message = exception.response.text
                raise RuntimeError(message) from exception

            raise exception from None
        return resp

    @staticmethod
    def _unwrap_enveloped_json(js: dict) -> dict:
        if isinstance(js, list):
            return js
        elif js is not None and js.get(ENVELOPED_JSON_TAG) is not None:
            return js[ENVELOPED_JSON_TAG]
        return js

    def get(self, endpoint, **kwargs) -> dict:
        r"""Sends a GET request. Returns :class:`dict` object.

        :param endpoint: endpoint to be appended to `client.host`.
        :param \*\*kwargs: Optional arguments that ``request`` takes.
        :rtype: dict
        """

        kwargs.setdefault("headers", self.headers)
        resp = self.session.get(f"{self.host}/{endpoint}", **kwargs)
        return self._unwrap_enveloped_json(self._raise_on_error(resp).json())

    def post(self, endpoint, data=None, json=None, dryrun=False, discard_response=False, **kwargs) -> dict:
        r"""Sends a POST request. Returns :class:`dict` object.

        :param endpoint: endpoint to be appended to `client.host`.
        :param data: (optional) Dictionary, list of tuples, bytes, or file-like
            object to send in the body of the :class:`Request`.
        :param json: (optional) json to send in the body of the :class:`Request`.
        :param \*\*kwargs: Optional arguments that ``request`` takes.
        :rtype: dict
        """

        if dryrun:
            headers = {**self.headers, **self.dryrun_header}
        else:
            headers = {**self.headers}
        kwargs.setdefault("headers", headers)

        resp = self.session.post(f"{self.host}/{endpoint}", data, filter_none(json), **kwargs)
        if discard_response:
            self._raise_on_error(resp)
            return None
        else:
            return self._unwrap_enveloped_json(self._raise_on_error(resp).json())

    def put(self, endpoint, data, **kwargs) -> dict:
        r"""Sends a PUT request. Returns :class:`dict` object.

        :param endpoint: endpoint to be appended to `client.host`.
        :param data: (optional) Dictionary, list of tuples, bytes, or file-like
            object to send in the body of the :class:`Request`.
        :param \*\*kwargs: Optional arguments that ``request`` takes.
        :rtype: dict
        """
        kwargs.setdefault("headers", self.headers)
        resp = self.session.put(f"{self.host}/{endpoint}", filter_none(data), **kwargs)
        return self._unwrap_enveloped_json(self._raise_on_error(resp).json())
